import {
  DecoratorParametersType,
  DecoratorType,
  NotSupportedDecoratorError,
  PropertyKeyType, Type,
} from '../core';

import {
  Class,
  Constructor,
  Field,
  Metadata,
  Method,
  Parameter,
} from '../reflect';

import {
  TypeUtil,
} from '../util';

import { Decorator } from './Decorator';

export class DecoratorHelper {

  public static register<Attribute = {}>(DecoratorClass: Type<Decorator<Attribute>>, attribute?: Attribute) {
    const annotation: Decorator<Attribute> = new DecoratorClass(attribute);
    const name: string = DecoratorClass.name;

    return function (...decoratorArgs: any[]) {
      return DecoratorHelper.registerAnnotation(name, annotation, decoratorArgs);
    };
  }

  public static create<Attribute = {}>(DecoratorClass: Type<Decorator<Attribute>>) {
    return function (attribute?: Attribute) {
      const annotation: Decorator<Attribute> = new DecoratorClass(attribute);
      const name: string = DecoratorClass.name;

      return function (...decoratorArgs: any[]) {
        return DecoratorHelper.registerAnnotation(name, annotation, decoratorArgs);
      };
    };
  }

  private static registerAnnotation(name: string, annotation: any, decoratorArgs: any[]) {
    const decoratorType: DecoratorType = DecoratorHelper.getDecoratorType(decoratorArgs);

    const [target, propertyKey, descriptorOrParameterIndex] = decoratorArgs;

    const clazz: Class = Class._defineClass(TypeUtil.getType(target));
    let field: Field;
    let method: Method;
    let parameter: Parameter | undefined;
    let cons: Constructor | null = null;

    switch (decoratorType) {
      case DecoratorType.CLASS:
        try {
          cons = clazz._defineConstructor(Metadata.getOwnParamTypes(target));
          clazz._addAnnotation(annotation);
          if (annotation instanceof Decorator && undefined !== annotation.classDecorator) {
            return annotation.classDecorator.call(annotation, target);
          }
        } catch (e) {
          if (e instanceof NotSupportedDecoratorError) {
            throw new NotSupportedDecoratorError(`Cannot apply @${name} decorator on Class.`);
          }
          throw e;
        }
        break;
      case DecoratorType.PROPERTY:
        try {
          field = clazz._defineField(propertyKey, Metadata.getOwnType(target, propertyKey));
          field._addAnnotation(annotation);
          if (annotation instanceof Decorator && undefined !== annotation.propertyDecorator) {
            return annotation.propertyDecorator.call(annotation, target, propertyKey);
          }
        } catch (e) {
          if (e instanceof NotSupportedDecoratorError) {
            throw new NotSupportedDecoratorError(`Cannot apply @${name} decorator on Property.`);
          }
          throw e;
        }
        break;
      case DecoratorType.METHOD:
        try {
          method = clazz._defineMethod(propertyKey,
                                     Metadata.getOwnParamTypes(target, propertyKey),
                                     Metadata.getOwnReturnType(target, propertyKey));
          method._addAnnotation(annotation);

          if (annotation instanceof Decorator && undefined !== annotation.methodDecorator) {
            return annotation.methodDecorator.call(annotation, target, propertyKey, descriptorOrParameterIndex);
          }
        } catch (e) {
          if (e instanceof NotSupportedDecoratorError) {
            throw new NotSupportedDecoratorError(`Cannot apply @${name} decorator on Method.`);
          }
          throw e;
        }
        break;
      case DecoratorType.PARAMETER:
        try {
          if (undefined === propertyKey) {
            cons = clazz.getConstructor();
            parameter = cons.getParameter(descriptorOrParameterIndex);
          } else {
            method = clazz._defineMethod(propertyKey,
                                       Metadata.getOwnParamTypes(target, propertyKey),
                                       Metadata.getOwnReturnType(target, propertyKey));
            parameter = method.getParameter(descriptorOrParameterIndex);
          }
          if (undefined !== parameter) {
            parameter._addAnnotation(annotation);
          }

          if (annotation instanceof Decorator && undefined !== annotation.parameterDecorator) {
            return annotation.parameterDecorator.call(annotation, target, propertyKey, descriptorOrParameterIndex);
          }
        } catch (e) {
          if (e instanceof NotSupportedDecoratorError) {
            throw new NotSupportedDecoratorError(`Cannot apply @${name} decorator on Parameter.`);
          }
          throw e;
        }
        break;
      default:
        throw new NotSupportedDecoratorError(`Cannot determine decorator[@${name}] type.`);
    }
  }

  /**
   *
   * @param {any[]} args
   * @returns {DecoratorType}
   */
  public static getDecoratorType(args: any[]): DecoratorType {
    const [, propertyKey, descriptor] = args;

    if (typeof descriptor === 'number') {
      return DecoratorType.PARAMETER;
    }

    if (propertyKey && descriptor === undefined || descriptor && (descriptor.get || descriptor.set)) {
      return DecoratorType.PROPERTY;
    }

    return (descriptor && descriptor.value) ? DecoratorType.METHOD : DecoratorType.CLASS;
  }

  /**
   *
   * @param target
   * @param {string} propertyKey
   * @returns {DecoratorParametersType}
   */
  public static decoratorArgs(target: any, propertyKey: PropertyKeyType): DecoratorParametersType {
    return [
      target,
      propertyKey,
      TypeUtil.descriptorOf(target, propertyKey),
    ];
  }
}
